#include <QtGui/QMouseEvent>
#include <cassert>
#include <cmath>
#include "GLWidget.h"
#include "GL/glu.h"

GLWidget::GLWidget(QWidget *parent):
    QGLWidget(parent),
    xmin(-2.),
    xmax(2.),
    ymin(-1.5),
    ymax(1.5),
    alpha(ALPHA0),
    beta(BETA0),
    mouseX(-1),     // Undefined
    mouseY(-1),     // Undefined
    quadric(0)
{
    setMouseTracking(true);
}

void GLWidget::initializeGL() {
    /*
    glDisable(GL_COLOR_MATERIAL);
    glEnable(GL_BLEND);
    glEnable(GL_POLYGON_SMOOTH);
    glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
    glClearColor(0., 0.1, 0.2, 1.);
    */

    glEnable(GL_DEPTH_TEST);
    glEnable(GL_LIGHTING);
    glDisable(GL_CULL_FACE);
    glEnable(GL_NORMALIZE);
    glDepthFunc(GL_LEQUAL);

    // Lighting
    glEnable(GL_LIGHT0);
    glLightModeli(GL_LIGHT_MODEL_TWO_SIDE, 1);

    GLfloat pos[4];
    pos[0] = 0.; pos[1] = 0.; pos[2] = 1.; pos[3] = 0.; // Directional light
    glLightfv(GL_LIGHT0, GL_POSITION, pos);

    GLfloat light[4];
    light[0] = 0.25; light[1] = 0.25; light[2] = 0.25; light[3] = 1.;
    glLightfv(GL_LIGHT0, GL_AMBIENT, light);

    light[0] = 1.; light[1] = 1.; light[2] = 1.; light[3] = 1.;
    glLightfv(GL_LIGHT0, GL_DIFFUSE, light);

    glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
}

void GLWidget::resizeGL(int w, int h) {
    glViewport(0, 0, w, h);
    glMatrixMode(GL_PROJECTION);
    glLoadIdentity();

    if (w == 0)
        w = 1;
    if (h == 0)
        h = 1;
    double aspect = (double) w / (double) h;
    assert(aspect > 0.);

    ymin = xmin / aspect;
    ymax = (-ymin);
    double depth = ymax - ymin;
    glOrtho(
        xmin, xmax,
        ymin, ymax,
        -10. * depth, 10. * depth
    );
    glMatrixMode(GL_MODELVIEW);
    glLoadIdentity();
}

void GLWidget::paintGL() {
    // Rotate the scene
    glMatrixMode(GL_MODELVIEW);
    glLoadIdentity();

    glEnable(GL_LIGHT0);        // Turn on the first light source

    // Rotate a model
    glRotatef(beta, 1., 0., 0.);
    glRotatef(alpha, 0., 1., 0.);

    render();           // Draw a scene graph

    // swapBuffers();   // Not needed - done automatically
}

void GLWidget::mousePressEvent(QMouseEvent * /* event */) {
    mouseX = (-1);      // Undefined
    mouseY = (-1);      // Undefined
}

void GLWidget::mouseReleaseEvent(QMouseEvent * /* event */) {
    mouseX = (-1);      // Undefined
    mouseY = (-1);      // Undefined
}

void GLWidget::mouseMoveEvent(QMouseEvent *event) {
    //... printf("%d, %d\n", event->x(), event->y());
    int x = event->x();
    int y = event->y();
    if ((event->buttons() & Qt::LeftButton) != 0) {
        if (mouseX >= 0) {
            int dx = x - mouseX;
            int dy = y - mouseY;
            if (beta > 100. || beta < (-100.))
                dx = -dx;

            alpha += dx * 0.1;
            if (fabs(alpha) > 360.)
                alpha -= ((int) alpha / 360) * 360.;

            beta += dy * 0.1;
            if (fabs(beta) > 360.)
                beta -= ((int) beta / 360) * 360.;

            repaint();
        }
        mouseX = x; mouseY = y;

    } else {
        mouseX = (-1);      // Undefined
        mouseY = (-1);      // Undefined
    }
}

void GLWidget::keyPressEvent(QKeyEvent* event) {
    switch (event->key()) {
    case Qt::Key_Escape:
    case Qt::Key_Q:
        close();
        break;
    case Qt::Key_Space:
        alpha = ALPHA0;
        beta = BETA0;
        repaint();
        break;
    default:
        event->ignore();
        break;
    }
}

void GLWidget::render() {
    GLfloat color[4], normal[3];

    glClearColor(0.2, 0.3, 0.5, 1.); // Background color: dark blue
    glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);

    // Construct a tetrahedron
    GLfloat side = sqrt(3.);
    GLfloat height = sqrt(2.);

    // Coordinates of tetrahedron vertices:
    GLfloat x0 = 0.,      y0 = (-height/4.), z0 = 1.;
    GLfloat x1 = side/2., y1 = y0,           z1 = (-0.5);
    GLfloat x2 = (-x1),   y2 = y0,           z2 = z1;
    GLfloat x3 = 0.,      y3 = height*3./4., z3 = 0.;

    // Draw a tetrahedron
    glBegin(GL_TRIANGLES);
        // Bottom face --------------------------------------
        color[0] = 0.2; color[1] = 1.; color[2] = 0.2;  // Green
        glMaterialfv(GL_FRONT_AND_BACK, GL_AMBIENT_AND_DIFFUSE, color);

        // Set a normal to the bottom face
        normal[0] = 0.;
        normal[1] = -1.;
        normal[2] = 0.;
        glNormal3fv(normal);

        // Define vertices of the face
        glVertex3f(x0, y0, z0);
        glVertex3f(x2, y2, z2);
        glVertex3f(x1, y1, z1);

        // Right side face --------------------------------------
        color[0] = 1.; color[1] = 0.2; color[2] = 0.2;  // Red
        glMaterialfv(GL_FRONT_AND_BACK, GL_AMBIENT_AND_DIFFUSE, color);

        // Set a normal to the right side face
        normal[0] = (x0 + x1 + x3) / 3.;
        normal[1] = (y0 + y1 + y3) / 3.;
        normal[2] = (z0 + z1 + z3) / 3.;
        glNormal3fv(normal);

        // Define vertices of the face
        glVertex3f(x0, y0, z0);
        glVertex3f(x1, y1, z1);
        glVertex3f(x3, y3, z3);

        // Back face --------------------------------------
        color[0] = 0.9; color[1] = 0.8; color[2] = 0.1; // Yellow
        glMaterialfv(GL_FRONT_AND_BACK, GL_AMBIENT_AND_DIFFUSE, color);

        // Define the normal to the face
        normal[0] = (x1 + x2 + x3) / 3.;
        normal[1] = (y1 + y2 + y3) / 3.;
        normal[2] = (z1 + z2 + z3) / 3.;
        glNormal3fv(normal);

        // Define vertices of the face
        glVertex3f(x1, y1, z1);
        glVertex3f(x2, y2, z2);
        glVertex3f(x3, y3, z3);

        // Left side face --------------------------------------
        color[0] = 0.2; color[1] = 0.2; color[2] = 1.;  // Blue
        glMaterialfv(GL_FRONT_AND_BACK, GL_AMBIENT_AND_DIFFUSE, color);

        // Define the normal to the face
        normal[0] = (x0 + x2 + x3) / 3.;
        normal[1] = (y0 + y2 + y3) / 3.;
        normal[2] = (z0 + z2 + z3) / 3.;
        glNormal3fv(normal);

        // Define vertices of the face
        glVertex3f(x2, y2, z2);
        glVertex3f(x0, y0, z0);
        glVertex3f(x3, y3, z3);
    glEnd();

    // Draw a sphere on top of tetrahedron
    if (quadric == 0) {
        quadric = gluNewQuadric();  // Create a Quadric object
        gluQuadricNormals(quadric, GLU_SMOOTH);
    }
    glTranslatef(x3, y3, z3);
    color[0] = 0.2; color[1] = 0.6; color[2] = 0.9;
    glMaterialfv(GL_FRONT_AND_BACK, GL_AMBIENT_AND_DIFFUSE, color);

    gluSphere(
        quadric,
        0.3,
        32,     // Num. slices (similar to lines of longitude)
        16      // Num. stacks (similar to lines of latitude)
    );
}
